import os
import cftime
import xarray as xr
import numpy as np
from scipy.spatial import cKDTree
from datetime import date, timedelta, datetime
import json
import pandas as pd

class NetCDFVertInterp:
    """Class for vertical interpolation of NetCDF data to station locations."""
    #constants
    A = 6378.137  # semieje mayor [km]
    E2 = 6.69437999014e-3  # excentricidad al cuadrado

    def __init__(self, expid, start_d, end_d, stn_lat, stn_lon):
        """Initialize the interpolator."""
        self.expid = expid
        self.start_d = start_d
        self.end_d = end_d
        self.stn_lat = stn_lat
        self.stn_lon = stn_lon
        self.n_neighbours = 4
        self.base_dir = f'/esarchive/exp/monarch/{self.expid}/'

        #optimization to avoid multiple os.listdir calls
        dir_content = os.listdir(f'/esarchive/exp/monarch/{self.expid}/')
        self.doms = [i for i in dir_content if i.startswith("d")]
        self.bool_check = any(i in self.doms for i in dir_content)
        self.r_or_g = [k for k in dir_content if k.startswith("g")]
        self.check_global_or_regional = any(k in self.r_or_g for k in dir_content)

    def _get_folder_path(self, variable, dom=None):
        """Return the folder path for a given variable."""
        if self.check_global_or_regional:
            return f'{self.base_dir}global/3hourly/{variable}/'
        else:
            if self.bool_check:
                return f'{self.base_dir}{dom}/hourly/{variable}/'
            return f'{self.base_dir}regional/hourly/{variable}/'

    def read_netcdfs(self, variables):
        """Read NetCDF files for multiple variables and return a dictionary."""

        dates = [
                    datetime.strftime(
                        self.start_d + timedelta(days=i), '%Y%m%d'
                    ) 
                    for i in range((self.end_d - self.start_d).days + 1)
                ]
        return {
            variable: self._process_variable(
                [f'{self._get_folder_path(variable)}/{variable}_{date}.nc' for date in dates], variable
            )
            for variable in variables
            if os.path.isdir(self._get_folder_path(variable))
        }


    @classmethod
    def lon_lat_to_cartesian(lon, lat):
        """Convert lon/lat coordinates to cartesian coordinates."""
        if lon.ndim == 1 and lat.ndim == 1:
            lon, lat = np.meshgrid(lon, lat)

        r_n = A / (np.sqrt(1 - E2 * (np.sin(lat_rad) ** 2)))
        x = r_n * np.cos(lat_rad) * np.cos(lon_rad)
        y = r_n * np.cos(lat_rad) * np.sin(lon_rad)
        z = r_n * (1 - E2) * np.sin(lat_rad)

        return x.flatten(), y.flatten(), z.flatten()


    @staticmethod
    def interpolate(source, target, dist, inds, variable):
        """Perform the interpolation."""
        w = 1.0 / dist ** 2
        sum_w = np.sum(w, axis=1)
        return [
            np.sum(w * source[variable][t].vlaues.flatten()[inds], axis=1) / sum_w
            for t in range(source['time'].shape[0])
        ]

    def save_data(self, output_folder, station_name, variables):
        """Save interpolated data to NetCDF files."""
        directory_path = os.path.join(output_folder, self.expid, str(self.start_d.year))
        os.makedirs(directory_path, exist_ok=True)
        data_dict = self.read_netcdfs(variables)

        for variable, data in data_dict.items():
            variable_dir = os.path.join(directory_path, variable)
            os.makedirs(variable_dir, exist_ok=True)
            data.to_csv(os.path.join(variable_dir, f"{variable}_{station_name}.csv"), index_label='time')
        
            
    def _process_variable(self, file_list, variable):
        dataframes = []
        lon_target, lat_target = np.meshgrid(self.stn_lon, self.stn_lat)
        for file_name in file_list:
            file_path = os.path.join(self._get_folder_path(variable), file_name)
            if os.path.isfile(file_path):
                with xr.open_dataset(file_path, decode_cf=False) as ds:
                    model_files = ds.isel(time=slice(None, -1))
                    #decode the true year of the file
                    time_units = model_files['time'].attrs['units']  # 'hours since 2018-01-01 00:00:00'
                    calendar = model_files['time'].attrs.get('calendar', 'standard')  # 'standard'
                    # Convierte los valores numéricos a fechas usando cftime
                    decoded_time = [str(t) for t in cftime.num2date(model_files['time'].values, units=time_units, calendar=calendar)]
                    model_files['time'] = ('time', pd.to_datetime(decoded_time))
                    
                    lat_source, lon_source = model_files.lat, model_files.lon
                    xs, ys, zs = self.lon_lat_to_cartesian(lon_source.values, lat_source.values)
                    xt, yt, zt = self.lon_lat_to_cartesian(lon_target.flatten(), lat_target.flatten())
                    tree = cKDTree(np.column_stack((xs, ys, zs)))
                    dist, inds = tree.query(np.column_stack((xt, yt, zt)), k=self.n_neighbours)

                    num_layers = model_files[variable].shape[1]
                    #print(f"Interpolating {variable} with {num_layers} layers.")
                    df_layers = pd.DataFrame()
                    #convert units 
                    convert_units = model_files[variable].attrs.get('units') == 'kgm-3'
                    for layer in range(num_layers):
                        interp_vals = self.interpolate(
                            model_files.isel(lev=layer),  # Asegúrate de usar 'lev'
                            lon_target,
                            dist,
                            inds,
                            variable
                        )
                        x = np.array([iv[0][0] * 1e9 if convert_units else iv[0][0] for iv in interp_vals])

                        df_layers[f"layer_{layer + 1}"] = x

                    # Manejo del índice temporal
                    df_layers.index = model_files['time'].values
                    dataframes.append(df_layers)
            else:
                print(f"File not found: '{file_name}'.")
        return pd.concat(dataframes, axis=0)

def load_stations(json_file):
    """Load station data from JSON file."""
    with open(json_file) as f:
        stn = json.load(f)
    return [(name, data['lat'], data['lon']) for name, data in stn.items()]
