import argparse
from datetime import date
from mpi4py import MPI
from Interpv03 import NetCDFVertInterp, load_stations
import variables_to_interpolate as vti
import time 
import multiprocessing as mp

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--expid', required=True)
    parser.add_argument('--start_date', required=True)
    parser.add_argument('--end_date', required=True)
    parser.add_argument('--output_dir', required=True)
    parser.add_argument('--station_file', required=True)
   # parser.add_argument('--variables', required=True)
    return parser.parse_args()

def process_stattion(args_tuple):
    """Auxiliary function to process each station with multiprocessing."""
    name, lat, lon, args, variables = args_tuple
    try: 
        start_time = time.time()
        interpolator = NetCDFVertInterp(
            expid=args.expid,
            start_d=date.fromisoformat(args.start_date),
            end_d=date.fromisoformat(args.end_date),
            stn_lat=lat,
            stn_lon=lon
        )
        interpolator.save_data(
            output_folder=args.output_dir,
            station_name=name,
            variables=variables
        )
        duration_time = time.time() - start_time
        return f"Processed station: {name} in {duration_time:.2f} seconds"

    except Exception as e:
        return f"Error processing station {name}: {str(e)}"


def main():
    args = parse_args()
    
    # Cargar estaciones
    stations = load_stations(args.station_file)
    
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    
    # Distribución de trabajo
    stations_per_task = len(stations) // size
    extra = len(stations) % size
    start_idx = rank * stations_per_task + min(rank, extra)
    end_idx = start_idx + stations_per_task + (1 if rank < extra else 0)
    my_stations = stations[start_idx:end_idx]
    
    #list od variables to process
    variables = vti.brown_carbon
    #build the tasks for each station
    tasks = [(name, lat, lon, args, variables) for name, lat, lon in my_stations]
    # Procesar cada estación asignada
    with mp.Pool(processes=mp.cpu_count()) as pool:
        results = pool.map(process_stattion, tasks)

    #reporting results by process
    for result in results:
        print(f'Rank {rank}: {result}')

    comm.Barrier()
    if rank == 0:
        print("All processing completed!")

if __name__ == "__main__":
    main()
